VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
END
Attribute VB_Name = "cLimitEdit"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
  Option Explicit
  ' demo project showing how to subclass the WM_CHAR and WM_PASTE
  ' messages to limit the text that can be entered into a textbox
  ' by Bryan Stafford of New Vision Software - newvision@mvps.org
  ' this demo is released into the public domain "as is" without
  ' warranty or guaranty of any kind.  In other words, use at
  ' your own risk.
  '
  '   Copyright  2000  Bryan Stafford/New Vision Software
  '
  ' the code in this demo is copyrighted and may only be used in
  ' accordance with the rules set forth in the "House Rules"
  ' section of the web page found at http://www.mvps.org/vbvision/
  
  ' this demo uses subclassing techniques so, if you don't already
  ' have it, you will want to get the Debug Object (Dbgwproc.dll)
  ' DLL from the MS website to allow stepping through subclassed
  ' code in break mode in the VB IDE.  before compiling this class,
  ' you will need to change the value of DEBUGWINDOWPROC in the
  ' Conditional Compiliation Arguements field on the Make tab of the
  ' Project Properties dialog to zero.
  
  
  ' the different types of limitations we can place on entries for the
  ' subclassed edit controls.
  Public Enum LIMIT_TYPE
    eDecimal = 0&
    eRealNumber
    eRealWithHyphen
    eAllowNegitiveNumbers = &H10000000
    eDoNotAllowContextMenu = &H40000000
    eDoNotAllowPaste = &H80000000
  End Enum
  
  ' if we are using the debug dll, we need a debug hook variable
  #If DEBUGWINDOWPROC Then
    Private m_oColProcHooks As Collection
  #End If
  
  ' original window proc addresses
  Private m_oColOldWindPorcs As Collection
  
Friend Function WindowProc(ByVal hWindow&, ByVal uMsg&, ByVal wParam&, ByVal lParam&) As Long
  
  Dim lpWindProcOld&, nLen&, sTemp$, eLimitType As LIMIT_TYPE

  lpWindProcOld = LoDWORD(m_oColOldWindPorcs(Hex$(hWindow)))


  Select Case uMsg
    Case WM_CHAR
      eLimitType = GetProp(hWindow, "nvLimitType")
      
      Select Case eLimitType And &HFF
        Case eDecimal
          Select Case wParam ' Ctl+C, Ctl+X, Ctl+Z, 0-9, Backspace
            Case 3, 24, 26, vbKey0 To vbKey9, vbKeyBack
              ' let these pass through....
        
            Case 46 ' "." <- only allow one decimal point
              If CheckForDuplicateChar(lpWindProcOld, hWindow, ".") Then
                Beep
                
                Exit Function
              End If
              
            Case 22 ' Ctl+V
              If (eLimitType And eDoNotAllowPaste) > 0 Then
                Beep
                
                Exit Function
              ElseIf ParsePastedString(lpWindProcOld, hWindow, eLimitType) = False Then
                Beep
                
                Exit Function
              End If
            
            Case 45 ' "-"
              If (eLimitType And eAllowNegitiveNumbers) = 0 Then
                Beep
                
                Exit Function
              ElseIf IsInvalidateNegitiveCharPos(lpWindProcOld, hWindow) Then
                Beep
                
                Exit Function
              End If
            
            Case Else
              Beep
              
              Exit Function
          End Select ' wParam
        
        
        Case eRealNumber
          Select Case wParam ' Ctl+C, Ctl+X, Ctl+Z, 0-9, Backspace
            Case 3, 24, 26, vbKey0 To vbKey9, vbKeyBack
              ' let these pass through....
              
            Case 22 ' Ctl+V
              If (eLimitType And eDoNotAllowPaste) > 0 Then
                Beep
                
                Exit Function
              ElseIf ParsePastedString(lpWindProcOld, hWindow, eLimitType) = False Then
                Beep
                
                Exit Function
              End If
                  
            Case 45 ' "-"
              If (eLimitType And eAllowNegitiveNumbers) = 0 Then
                Beep
                
                Exit Function
              ElseIf IsInvalidateNegitiveCharPos(lpWindProcOld, hWindow) Then
                Beep
                
                Exit Function
              End If
                  
            Case Else
              Beep
              
              Exit Function
          End Select ' wParam
        

        Case eRealWithHyphen ' use this for social security numbers....
          Select Case wParam ' Ctl+C, Ctl+X, Ctl+Z, "-",  0-9, Backspace
            Case 3, 24, 26, 45, vbKey0 To vbKey9, vbKeyBack
              ' let these pass through....
              
            Case 22 ' Ctl+V
              If (eLimitType And eDoNotAllowPaste) > 0 Then
                Beep
                
                Exit Function
              ElseIf ParsePastedString(lpWindProcOld, hWindow, eLimitType) = False Then
                Beep
                
                Exit Function
              End If
            
            Case Else
              Beep
              
              Exit Function
          End Select ' wParam

      End Select ' GetProp(hWindow, "nvLimitType")
      
    Case WM_PASTE
      eLimitType = GetProp(hWindow, "nvLimitType")
      
      ' see if we are allowing the user to paste into this control
      If (eLimitType And eDoNotAllowPaste) <> 0 Then
        Beep
        
        Exit Function
      
      ElseIf ParsePastedString(lpWindProcOld, hWindow, eLimitType) = False Then
        ' if we are allowing the user to paste, see if there are any invalid chars in the string on the clipboard
        Beep
        
        Exit Function
      End If
      
      
    Case WM_CONTEXTMENU ' see whether or not we are allowing the right click menu....
      If (GetProp(hWindow, "nvLimitType") And eDoNotAllowContextMenu) <> 0 Then
        Beep
        
        Exit Function
      End If
       
  End Select ' uMsg
  
  ' pass the messages on to VB
  WindowProc = CallWindowProc(lpWindProcOld, hWindow, uMsg, wParam, lParam)

End Function

Private Function ParsePastedString(ByVal lpWindProcOld&, ByVal hWnd&, ByVal eLimitType As LIMIT_TYPE) As Boolean
  ' parse a pasted string to make sure it and any unselected text left in the control will conform to the desired format
  
  Dim i&, sClipboard$, nChar&, bDecimalFound As Boolean, bInvalidNegitivePos As Boolean, eContentType As LIMIT_TYPE

  eContentType = (eLimitType And &HFF)
  
  sClipboard = Clipboard.GetText

  If sClipboard <> vbNullString Then
    ' assume success....
    ParsePastedString = True

    Select Case eContentType
      Case eDecimal: bDecimalFound = CheckForDuplicateChar(lpWindProcOld, hWnd, ".")
    End Select
    
    If (eLimitType And eAllowNegitiveNumbers) <> 0 Then
      bInvalidNegitivePos = IsInvalidateNegitiveCharPos(lpWindProcOld, hWnd)
    End If
    
    For i = 1 To Len(sClipboard)
      nChar = Asc(Mid$(sClipboard, i, 1))
      
      Select Case eContentType
        Case eDecimal
          Select Case nChar
            Case vbKey0 To vbKey9
              ' chars OK, do nothing
            
            Case 46 ' "." <- only allow one decimal point
              If bDecimalFound Then
                ParsePastedString = False
                
                Exit For
              Else
                bDecimalFound = True
              End If
            
            Case 45 ' "-"
              If (eLimitType And eAllowNegitiveNumbers) = 0 Then
                ParsePastedString = False
                
                Exit Function
              Else ' make sure this is the first char in the string
                   ' this really *should* check to see how negitive
                   ' numbers are formated in the current locale....
                If (bInvalidNegitivePos = True) Or (i > 1) Then
                  ParsePastedString = False
                  
                  Exit Function
                End If
              End If
            
            Case Else
              ParsePastedString = False
              
              Exit For
            
          End Select
          
        Case eRealNumber
          Select Case nChar
            Case vbKey0 To vbKey9
              ' chars OK, do nothing
            
            Case 45 ' "-"
              If (eLimitType And eAllowNegitiveNumbers) = 0 Then
                ParsePastedString = False
                
                Exit Function
              Else ' make sure this is the first char in the string
                   ' this really *should* check to see how negitive
                   ' numbers are formated in the current locale....
                If (bInvalidNegitivePos = True) Or (i > 1) Then
                  ParsePastedString = False
                  
                  Exit Function
                End If
              End If
            
            Case Else
              ParsePastedString = False
              
              Exit For
            
          End Select
      
        Case eRealWithHyphen
          Select Case nChar
            Case vbKey0 To vbKey9, 45 ' "-"
              ' chars OK, do nothing
                        
            Case Else
              ParsePastedString = False
              
              Exit For
            
          End Select

      End Select
    Next
  End If
  
End Function

Private Function CheckForDuplicateChar(ByVal lpWindProcOld&, ByVal hWnd&, ByVal sChar$) As Boolean
  ' returns true if a char matching the value in sChar$ already exists in the string
  
  Dim nLen&, sText$, nSelStart&, nSelEnd&

  nLen = CallWindowProc(lpWindProcOld, hWnd, WM_GETTEXTLENGTH, 0&, 0&)

  If nLen Then
    nLen = nLen + 1
    sText = String$(nLen, 0)

    nLen = CallWindowProclParamStr(lpWindProcOld, hWnd, WM_GETTEXT, nLen, sText)
    If nLen Then
      sText = left$(sText, nLen)
      
      Call CallWindowProc(lpWindProcOld, hWnd, EM_GETSEL, VarPtr(nSelStart), VarPtr(nSelEnd))
      
      If nSelStart <> nSelEnd Then
        sText = left$(sText, nSelStart) & right$(sText, nLen - nSelEnd)
      End If
      
      If InStr(sText, sChar) Then CheckForDuplicateChar = True
    End If
  End If

End Function

Private Function IsInvalidateNegitiveCharPos(ByVal lpWindProcOld&, ByVal hWnd&) As Boolean
  ' returns true if a negitive sign already exists in the edit control or if the
  ' caret position in the control is anywhere other than at the begining of the text
  
  Dim nSelStart&, nSelEnd&

  Call CallWindowProc(lpWindProcOld, hWnd, EM_GETSEL, VarPtr(nSelStart), VarPtr(nSelEnd))
  
  If CheckForDuplicateChar(lpWindProcOld, hWnd, "-") Then
    IsInvalidateNegitiveCharPos = True
  ElseIf nSelStart <> 0 Then
    IsInvalidateNegitiveCharPos = True
  End If
      
End Function

Public Sub AddLimitedItem(ByVal hTarget&, Optional ByVal eEntryType As LIMIT_TYPE = eDecimal Or eAllowNegitiveNumbers)

  If SubClass(hTarget) Then
    ' if a limit type is already set, remove it and then set the new type
    If GetProp(hTarget, "nvLimitType") Then Call RemoveProp(hTarget, "nvLimitType")
    
    Call SetProp(hTarget, ByVal "nvLimitType", eEntryType)
  End If

End Sub

Private Function SubClass(ByVal hWnd&) As Boolean
  ' this routine subclasses the window
  
  ' first check with the OS to make sure the hWnd is a valid window
  If IsWindow(hWnd) Then
    ' we store a pointer to this instance of the class in an extra property
    ' we add to the window using the SetProp API function.  if the property exists,
    ' this window is already subclassed and we don't want to overwrite the property
    If GetProp(hWnd, "nvLimitEdit") = 0 Then
    
      ' store a pointer to this instance of the class object in an added property
      If SetProp(hWnd, ByVal "nvLimitEdit", ObjPtr(Me)) Then
        ' if the SetProp call succeeded, subclass the window...
        
        If m_oColOldWindPorcs Is Nothing Then Set m_oColOldWindPorcs = New Collection
        
        ' we use a conditional compilation argument to tell whether or not we should use the
        ' debug dll.  when you want to compile a release, set the DEBUGWINDOWPROC variable in
        ' the project properties dialog to zero.
        #If DEBUGWINDOWPROC Then
          Dim lpWindProc&, oDWPHook As WindowProcHook
          
          ' this is the procedure for setting a debug hook...
          Set oDWPHook = CreateWindowProcHook
          With oDWPHook
          
            ' our windowproc is in the MLimitEdit BAS module
            .SetMainProc AddressOf MLimitEdit.LimitEditWindProc
            
            ' replace the windowproc for this window with ours
            lpWindProc = SetWindowLong(hWnd, GWL_WNDPROC, .ProcAddress)
            
            ' store the original proc address and the hWnd in our collection.
            ' use the hWnd as the key for the item
            m_oColOldWindPorcs.Add MAKEDOUBLE(lpWindProc, hWnd), Hex$(hWnd)
            
            .SetDebugProc lpWindProc
          End With
          
          ' store our debug hook in the collection of debug hooks
          If m_oColProcHooks Is Nothing Then Set m_oColProcHooks = New Collection
          
          m_oColProcHooks.Add oDWPHook, Hex$(hWnd)
          
          Set oDWPHook = Nothing
          
        #Else
          ' not using the debug dll so just replace the windowproc for this window with ours.
          ' our windowproc is in the MLimitEdit BAS module
          m_oColOldWindPorcs.Add MAKEDOUBLE(SetWindowLong(hWnd, GWL_WNDPROC, AddressOf MLimitEdit.LimitEditWindProc), hWnd), Hex$(hWnd)
        #End If
    
        SubClass = True
      End If
    Else
      SubClass = True
    End If
  End If
  
End Function

Public Sub RemoveLimits(ByVal hTarget&)

  If hTarget <> (-1) Then UnSubClass hTarget
  
End Sub

Private Sub UnSubClass(ByVal hWnd&)
  ' this routine un-subclasses the window
  
  Dim curWindowData@, lpOldProc&, vhWnd As Variant
  
  If hWnd = (-1) Then
    For Each vhWnd In m_oColOldWindPorcs
      curWindowData = vhWnd
      
      hWnd = HiDWORD(curWindowData)
      lpOldProc = LoDWORD(curWindowData)
      
      ' make sure the window still exists....
      If IsWindow(hWnd) Then
        ' if we have an address for the old windowproc, return the proc to the original state...
        
        ' un-subclass the window
        SetWindowLong hWnd, GWL_WNDPROC, lpOldProc
        
        ' remove the added properties
        RemoveProp hWnd, "nvLimitEdit"
        RemoveProp hWnd, "nvLimitType"
        
        ' set the variables to zero to avoid any misshaps
        m_oColOldWindPorcs.Remove Hex$(hWnd)
        
        #If DEBUGWINDOWPROC Then
          m_oColProcHooks.Remove Hex$(hWnd)
        #End If
      End If
    Next
    
  Else
    ' make sure the window still exists....
    If IsWindow(hWnd) Then
      ' if we have an address for the old windowproc, return the proc to the original state...
      curWindowData = m_oColOldWindPorcs(Hex$(hWnd))
      lpOldProc = LoDWORD(curWindowData)
      
      ' un-subclass the window
      SetWindowLong hWnd, GWL_WNDPROC, lpOldProc
      
      ' remove the added properties
      RemoveProp hWnd, "nvLimitEdit"
      RemoveProp hWnd, "nvLimitType"
      
      ' set the variables to zero to avoid any misshaps
      m_oColOldWindPorcs.Remove Hex$(hWnd)
      
      #If DEBUGWINDOWPROC Then
        m_oColProcHooks.Remove Hex$(hWnd)
      #End If
    End If
  End If
  
  If m_oColOldWindPorcs.Count = 0 Then
    ' if we are using the debug dll, destroy the instance
    #If DEBUGWINDOWPROC Then
      Set m_oColProcHooks = Nothing
    #End If
  
    Set m_oColOldWindPorcs = Nothing
  End If
  
End Sub

Private Sub Class_Terminate()
  ' make sure we clean up after ourselves....
  If Not m_oColOldWindPorcs Is Nothing Then UnSubClass (-1)
End Sub
